@@ -4,7 +4,7 @@ import numpy as np
 import os
 from numba import njit, prange
 
-from .maxcut_tfim_util import opencl_context
+from .maxcut_tfim_util import get_cut, init_theta, init_thresholds, low_width, maxcut_hamming_cdf, opencl_context, probability_by_hamming_weight
 
 IS_OPENCL_AVAILABLE = True
 try:
@@ -13,89 +13,6 @@ except ImportError:
     IS_OPENCL_AVAILABLE = False
 
 
-@njit
-def probability_by_hamming_weight(J, h, z, theta, t, n_qubits):
-    bias = np.empty(n_qubits - 1, dtype=np.float32)
-
-    # critical angle
-    theta_c = np.arcsin(max(-1.0, min(1.0, abs(h) / (z * J))))
-
-    p = (
-        pow(2.0, abs(J / h) - 1.0)
-        * (1.0 + np.sin(theta - theta_c) * np.cos(1.5 * np.pi * J * t + theta) / (1.0 + np.sqrt(t)))
-        - 0.5
-    )
-
-    if (p * n_qubits) >= 1024:
-        return bias
-
-    tot_n = 1.0 + 1.0 / pow(2.0, p * n_qubits)
-    factor = pow(2.0, -p)
-    n = 1.0
-    for q in range(1, n_qubits):
-        n *= factor
-        bias[q - 1] = n
-        tot_n += n
-    bias /= tot_n
-
-    if J > 0.0:
-        return bias[::-1]
-
-    return bias
-
-
-@njit(parallel=True)
-def maxcut_hamming_cdf(n_qubits, J_func, degrees, quality, hamming_prob):
-    if n_qubits < 2:
-        hamming_prob.fill(0.0)
-        return
-
-    n_steps = 1 << quality
-    delta_t = 1.0 / n_steps
-    tot_t = 2.0 * n_steps * delta_t
-    h_mult = 2.0 / tot_t
-    n_bias = n_qubits - 1
-
-    theta = np.empty(n_qubits, dtype=np.float32)
-    for q in range(n_qubits):
-        J = J_func[q]
-        z = degrees[q]
-        theta[q] = np.arcsin(
-            max(
-                -1.0,
-                min(
-                    1.0,
-                    (1.0 if J > 0.0 else -1.0) if np.isclose(abs(z * J), 0.0) else (abs(h_mult) / (z * J)),
-                ),
-            )
-        )
-
-    for qc in prange(n_qubits, n_steps * n_qubits):
-        step = qc // n_qubits
-        q = qc % n_qubits
-        J_eff = J_func[q]
-        if np.isclose(abs(J_eff), 0.0):
-            continue
-        z = degrees[q]
-        theta_eff = theta[q]
-        t = step * delta_t
-        tm1 = (step - 1) * delta_t
-        h_t = h_mult * (tot_t - t)
-        bias = probability_by_hamming_weight(J_eff, h_t, z, theta_eff, t, n_qubits)
-        last_bias = probability_by_hamming_weight(J_eff, h_t, z, theta_eff, tm1, n_qubits)
-        for i in range(n_bias):
-            hamming_prob[i] += bias[i] - last_bias[i]
-
-    tot_prob = hamming_prob.sum()
-    hamming_prob /= tot_prob
-
-    tot_prob = 0.0
-    for i in range(n_bias):
-        tot_prob += hamming_prob[i]
-        hamming_prob[i] = tot_prob
-    hamming_prob[-1] = 2.0
-
-
 @njit
 def binary_search(l, t):
   left = 0
@@ -237,9 +154,8 @@ def init_J_and_z(G_data, G_rows, G_cols):
         degree = end - start
         val = G_data[start:end].sum()
 
-        J = -val / degree if degree > 0 else 0
         degrees[r] += degree
-        J_eff[r] += J
+        J_eff[r] += val
 
         # Column sum
         for idx in range(start, end):
@@ -260,43 +176,33 @@ def init_J_and_z(G_data, G_rows, G_cols):
 
 
 @njit
-def init_thresholds(n_qubits):
-    n_bias = n_qubits - 1
-    thresholds = np.empty(n_bias, dtype=np.float32)
-    tot_prob = 0
-    p = 1.0
-    if n_qubits & 1:
-        q = n_qubits // 2
-        thresholds[q - 1] = p
-        tot_prob = p
-        p /= 2
-    for q in range(1, n_qubits // 2):
-        thresholds[q - 1] = p
-        thresholds[n_bias - q] = p
-        tot_prob += 2 * p
-        p /= 2
-    thresholds /= tot_prob
+def cpu_footer(shots, quality, n_qubits, G_data, G_rows, G_cols, nodes):
+    J_eff, degrees = init_J_and_z(G_data, G_rows, G_cols)
+    hamming_prob = init_thresholds(n_qubits)
 
-    return thresholds
+    maxcut_hamming_cdf(n_qubits, J_eff, degrees, quality, hamming_prob)
+
+    best_solution, best_value = sample_for_solution(G_data, G_rows, G_cols, shots, hamming_prob, J_eff)
+
+    bit_string, l, r = get_cut(best_solution, nodes)
+
+    return bit_string, best_value, (l, r)
 
 
-@njit(parallel=True)
-def init_theta(delta_t, tot_t, h_mult, n_qubits, J_eff, degrees):
-    theta = np.empty(n_qubits, dtype=np.float32)
-    for q in prange(n_qubits):
-        J = J_eff[q]
-        z = degrees[q]
-        theta[q] = np.arcsin(
-            max(
-                -1.0,
-                min(
-                    1.0,
-                    (1.0 if J > 0.0 else -1.0) if np.isclose(abs(z * J), 0.0) else (abs(h_mult) / (z * J)),
-                ),
-            )
-        )
+@njit
+def gpu_footer(shots, n_qubits, G_data, G_rows, G_cols, J_eff, hamming_prob, nodes):
+    hamming_prob /= hamming_prob.sum()
+    tot_prob = 0.0
+    for i in range(n_qubits - 1):
+        tot_prob += hamming_prob[i]
+        hamming_prob[i] = tot_prob
+    hamming_prob[-1] = 2.0
 
-    return theta
+    best_solution, best_value = sample_for_solution(G_data, G_rows, G_cols, shots, hamming_prob, J_eff)
+
+    bit_string, l, r = get_cut(best_solution, nodes)
+
+    return bit_string, best_value, (l, r)
 
 
 def to_scipy_sparse_upper_triangular(G, nodes, n_nodes):
@@ -329,18 +235,7 @@ def maxcut_tfim_sparse(
         G_m = G
 
     if n_qubits < 3:
-        if n_qubits == 0:
-            return "", 0, ([], [])
-
-        if n_qubits == 1:
-            return "0", 0, (nodes, [])
-
-        if n_qubits == 2:
-            weight = G_m[0, 1]
-            if weight < 0.0:
-                return "00", 0, (nodes, [])
-
-            return "01", weight, ([nodes[0]], [nodes[1]])
+        return low_width(G_m, nodes, n_qubits)
 
     if quality is None:
         quality = 8
@@ -352,61 +247,42 @@ def maxcut_tfim_sparse(
     n_steps = 1 << quality
     grid_size = n_steps * n_qubits
 
+    if not (IS_OPENCL_AVAILABLE and grid_size >= 128):
+        return cpu_footer(shots, quality, n_qubits, G_m.data, G_m.indptr, G_m.indices, nodes)
+
     J_eff, degrees = init_J_and_z(G_m.data, G_m.indptr, G_m.indices)
     hamming_prob = init_thresholds(n_qubits)
 
-    if IS_OPENCL_AVAILABLE and grid_size >= 128:
-        delta_t = 1.0 / n_steps
-        tot_t = 2.0 * n_steps * delta_t
-        h_mult = 2.0 / tot_t
-        theta = init_theta(delta_t, tot_t, h_mult, n_qubits, J_eff, degrees)
-        args = np.empty(3, dtype=np.float32)
-        args[0] = delta_t
-        args[1] = tot_t
-        args[2] = h_mult
+    delta_t = 1.0 / n_steps
+    tot_t = 2.0 * n_steps * delta_t
+    h_mult = 2.0 / tot_t
+    theta = init_theta(delta_t, tot_t, h_mult, n_qubits, J_eff, degrees)
+    args = np.empty(3, dtype=np.float32)
+    args[0] = delta_t
+    args[1] = tot_t
+    args[2] = h_mult
 
-        # Warp size is 32:
-        group_size = n_qubits - 1
-        if group_size > 256:
-            group_size = 256
-        grid_dim = n_steps * n_qubits * group_size
+    # Warp size is 32:
+    group_size = n_qubits - 1
+    if group_size > 256:
+        group_size = 256
+    grid_dim = n_steps * n_qubits * group_size
 
-        # Move to GPU
-        mf = cl.mem_flags
-        args_buf = cl.Buffer(opencl_context.ctx, mf.READ_ONLY | mf.COPY_HOST_PTR, hostbuf=args)
-        J_buf = cl.Buffer(opencl_context.ctx, mf.READ_ONLY | mf.COPY_HOST_PTR, hostbuf=J_eff)
-        deg_buf = cl.Buffer(opencl_context.ctx, mf.READ_ONLY | mf.COPY_HOST_PTR, hostbuf=degrees)
-        theta_buf = cl.Buffer(opencl_context.ctx, mf.READ_ONLY | mf.COPY_HOST_PTR, hostbuf=theta)
-        ham_buf = cl.Buffer(opencl_context.ctx, mf.READ_WRITE | mf.COPY_HOST_PTR, hostbuf=hamming_prob)
+    # Move to GPU
+    mf = cl.mem_flags
+    args_buf = cl.Buffer(opencl_context.ctx, mf.READ_ONLY | mf.COPY_HOST_PTR, hostbuf=args)
+    J_buf = cl.Buffer(opencl_context.ctx, mf.READ_ONLY | mf.COPY_HOST_PTR, hostbuf=J_eff)
+    deg_buf = cl.Buffer(opencl_context.ctx, mf.READ_ONLY | mf.COPY_HOST_PTR, hostbuf=degrees)
+    theta_buf = cl.Buffer(opencl_context.ctx, mf.READ_ONLY | mf.COPY_HOST_PTR, hostbuf=theta)
+    ham_buf = cl.Buffer(opencl_context.ctx, mf.READ_WRITE | mf.COPY_HOST_PTR, hostbuf=hamming_prob)
 
-        # Kernel execution
-        opencl_context.maxcut_hamming_cdf_kernel(
-            opencl_context.queue, (grid_dim,), (group_size,),
-            np.int32(n_qubits), deg_buf, args_buf, J_buf, theta_buf, ham_buf
-        )
+    # Kernel execution
+    opencl_context.maxcut_hamming_cdf_kernel(
+        opencl_context.queue, (grid_dim,), (group_size,),
+        np.int32(n_qubits), deg_buf, args_buf, J_buf, theta_buf, ham_buf
+    )
 
-        # Fetch results
-        cl.enqueue_copy(opencl_context.queue, hamming_prob, ham_buf)
+    # Fetch results
+    cl.enqueue_copy(opencl_context.queue, hamming_prob, ham_buf)
 
-        hamming_prob /= hamming_prob.sum()
-        tot_prob = 0.0
-        for i in range(n_qubits - 1):
-            tot_prob += hamming_prob[i]
-            hamming_prob[i] = tot_prob
-        hamming_prob[-1] = 2.0
-    else:
-        maxcut_hamming_cdf(n_qubits, J_eff, degrees, quality, hamming_prob)
-
-    best_solution, best_value = sample_for_solution(G_m.data, G_m.indptr, G_m.indices, shots, hamming_prob, J_eff)
-
-    bit_string = ""
-    l, r = [], []
-    for i in range(len(best_solution)):
-        if best_solution[i]:
-            bit_string += "1"
-            r.append(nodes[i])
-        else:
-            bit_string += "0"
-            l.append(nodes[i])
-
-    return bit_string, best_value, (l, r)
+    return gpu_footer(shots, n_qubits, G_m.data, G_m.indptr, G_m.indices, J_eff, hamming_prob, nodes)
