@@ -42,8 +42,8 @@ def bootstrap_worker(theta, G_m, indices):
 
 @njit(parallel=True)
 def bootstrap(theta, G_m, k, indices_array):
-    n = len(indices_array) // k
-    energies = np.empty(n, dtype=np.float64)
+    n = theta.shape[0]
+    energies = np.empty(n, dtype=np.float32)
     for i in prange(n):
         j = i * k
         energies[i] = bootstrap_worker(theta, G_m, indices_array[j : j + k])
@@ -56,7 +56,7 @@ def int_to_bitstring(integer, length):
     return (bin(integer)[2:].zfill(length))[::-1]
 
 
-def spin_glass_solver(G, quality=6, shots=None, best_guess=None):
+def spin_glass_solver(G, quality=None, shots=None, best_guess=None):
     nodes = None
     n_qubits = 0
     G_m = None
@@ -83,6 +83,9 @@ def spin_glass_solver(G, quality=6, shots=None, best_guess=None):
 
             return "01", weight, ([nodes[0]], [nodes[1]]), -weight
 
+    if quality is None:
+        quality = 6
+
     bitstring = ""
     if isinstance(best_guess, str):
         bitstring = best_guess
@@ -92,7 +95,7 @@ def spin_glass_solver(G, quality=6, shots=None, best_guess=None):
         bitstring = "".join(["1" if b else "0" for b in best_guess])
     else:
         bitstring, _, _ = maxcut_tfim(G_m, quality=quality, shots=shots)
-    best_theta = [b == "1" for b in list(bitstring)]
+    best_theta = np.array([b == "1" for b in list(bitstring)], dtype=np.bool_)
 
     min_energy = compute_energy(best_theta, G_m)
     improved = True
